from typing import Any

from eth_typing import ChecksumAddress as EthChecksumAddress
from pydantic import validate_call
from web3.contract.contract import Contract, ContractFunction

from .client import SdkClient
from .interfaces import IJPYC
from .utils.addresses import (
    get_proxy_address,
)
from .utils.artifacts import (
    get_artifacts,
    resolve_artifacts_file_path,
)
from .utils.chains import SUPPORTED_CHAINS
from .utils.constants import SIGN_MIDDLEWARE
from .utils.currencies import (
    remove_decimals,
    restore_decimals,
)
from .utils.errors import (
    AccountNotInitialized,
    TransactionFailed,
    TransactionSimulationFailed,
)
from .utils.types import ContractVersion, TransactionArgs
from .utils.validators import Bytes32, ChecksumAddress, Uint8, Uint256


class JPYC(IJPYC):
    """Implementation of IJPYC."""

    def __init__(
        self,
        client: SdkClient,
        contract_version: ContractVersion = "2",
        contract_address: EthChecksumAddress | None = None,
    ) -> None:
        """Constructor that initializes JPYC client.

        Notes:
            - If `client` parameter is configured to use localhost network,\
            this deploys JPYC contracts to localhost network, initializes it,\
            and sets its address to `address` attribute.
            - If `contract_address` is supplied,\
            this configures contract instance with that address.

        Args:
            client (SdkClient): Configured SDK client
            contract_version (ContractVersion): Contract version
            contract_address (EthChecksumAddress, optional): Contract address
        """
        if (
            client.w3.eth.chain_id == SUPPORTED_CHAINS["localhost"]["devnet"]["id"]
            and contract_address is None
        ):
            address = self.__deploy_contract(
                client=client,
                contract_version=contract_version,
            )
            contract = self.__get_contract(
                client=client,
                contract_address=address,
                contract_version=contract_version,
            )
            self.__initialize_contract(
                client=client,
                contract=contract,
            )
        else:
            address = (
                contract_address
                if contract_address is not None
                else get_proxy_address(contract_version=contract_version)
            )
            contract = self.__get_contract(
                client=client,
                contract_address=address,
                contract_version=contract_version,
            )

        self.client = client
        """ISdkClient: Configured SDK client"""
        self.contract = contract
        """Contract: Configured contract instance"""

    ##################
    # Helper methods #
    ##################

    @staticmethod
    def __deploy_contract(
        client: SdkClient,
        contract_version: ContractVersion = "2",
    ) -> ChecksumAddress:
        """Deploy contracts to the configured network.

        Note:
            This helper method is mainly for development purposes. \
            Please use this method to deploy contracts to localhost network.

        Args:
            client (SdkClient): Configured SDK client
            contract_version (ContractVersion): Contract version

        Returns:
            ChecksumAddress: Address of the deployed contracts
        """
        file_path = resolve_artifacts_file_path(contract_version=contract_version)
        contract = client.w3.eth.contract(
            abi=get_artifacts(file_path, "abi"),
            bytecode=get_artifacts(file_path, "bytecode"),
        )
        tx_hash = contract.constructor().transact()

        return client.w3.eth.wait_for_transaction_receipt(tx_hash).contractAddress  # type: ignore[attr-defined]

    @staticmethod
    def __get_contract(
        client: SdkClient,
        contract_address: ChecksumAddress,
        contract_version: ContractVersion = "2",
    ) -> Contract:
        """Get contract instance from the configured network.

        Args:
            client (SdkClient): Configured SDK client
            contract_address (ChecksumAddress): Contract address
            contract_version (ContractVersion): Contract version

        Returns:
            Contract: Address of the deployed contracts
        """
        return client.w3.eth.contract(  # type: ignore[call-overload]
            address=contract_address,
            abi=get_artifacts(
                file_path=resolve_artifacts_file_path(
                    contract_version=contract_version
                ),
                artifact_type="abi",
            ),
        )

    @staticmethod
    def __initialize_contract(
        client: SdkClient,
        contract: Contract,
    ) -> None:
        """Initialize contract.

        Args:
            client (SdkClient): Configured SDK client
            contract (Contract): Configured contract instance
        """
        owner_address = client.get_account_address()
        contract.functions.initialize(
            "JPY Coin",
            "JPYC",
            "Yen",
            18,
            owner_address,
            owner_address,
            owner_address,
            owner_address,
            owner_address,
        ).transact()

    def __account_initialized(self) -> None:
        """Checks if account is initialized.

        Note:
            An account must be set to web3 instance to send transactions.

        Raises:
            AccountNotInitialized: If account is not initialized
        """
        if SIGN_MIDDLEWARE not in self.client.w3.middleware_onion:
            raise AccountNotInitialized()

    @staticmethod
    def __simulate_transaction(
        contract_func: ContractFunction,
        func_args: dict[str, object],
    ) -> None:
        """Simulates a transaction locally.

        Note:
            This method should be called before sending actual transactions.

        Args:
            contract_func (ContractFunction): Contract function
            func_args (dict[str, object]): Arguments of contract function

        Raises:
            TransactionSimulationFailed: If transaction simulation fails
        """
        try:
            contract_func(**func_args).call()
        except Exception as e:
            raise TransactionSimulationFailed(str(e))

    @staticmethod
    def __send_transaction(
        contract_func: ContractFunction,
        func_args: dict[str, object],
    ) -> Any:
        """Sends a transaction to blockchain.

        Args:
            contract_func (ContractFunction): Contract function
            func_args (dict[str, object]): Arguments of contract function

        Returns:
            Any: Response from the contract function

        Raises:
            TransactionFailed: If transaction fails
        """
        try:
            return contract_func(**func_args).transact()
        except Exception as e:
            raise TransactionFailed(str(e))

    def __transact(
        self, contract_func: ContractFunction, func_args: dict[str, object]
    ) -> Any:
        """Helper method to prepare & send a transaction in one method.

        Args:
            contract_func (ContractFunction): Contract function
            func_args (dict[str, object]): Arguments of contract function

        Returns:
            Any: Response from the contract function

        Raises:
            AccountNotInitialized: If account is not initialized
            TransactionSimulationFailed: If transaction simulation fails
            TransactionFailed: If transaction fails
        """

        self.__account_initialized()
        self.__simulate_transaction(
            contract_func,
            func_args,
        )
        return self.__send_transaction(
            contract_func,
            func_args,
        )

    ##################
    # View functions #
    ##################

    @validate_call
    def is_minter(self, account: ChecksumAddress) -> bool:
        return self.contract.functions.isMinter(account).call()

    @restore_decimals
    @validate_call
    def minter_allowance(self, minter: ChecksumAddress) -> Uint256:
        return self.contract.functions.minterAllowance(minter).call()

    @restore_decimals
    def total_supply(self) -> Uint256:
        return self.contract.functions.totalSupply().call()

    @restore_decimals
    @validate_call
    def balance_of(self, account: ChecksumAddress) -> Uint256:
        return self.contract.functions.balanceOf(account).call()

    @restore_decimals
    @validate_call
    def allowance(self, owner: ChecksumAddress, spender: ChecksumAddress) -> Uint256:
        return self.contract.functions.allowance(owner, spender).call()

    @validate_call
    def nonces(self, owner: ChecksumAddress) -> Uint256:
        return self.contract.functions.nonces(owner).call()

    ######################
    # Mutation functions #
    ######################

    @validate_call
    def configure_minter(
        self, minter: ChecksumAddress, minter_allowed_amount: Uint256
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.configureMinter,
            "func_args": {
                "minter": minter,
                "minterAllowedAmount": remove_decimals(minter_allowed_amount),
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def mint(self, to: ChecksumAddress, amount: Uint256) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.mint,
            "func_args": {
                "_to": to,
                "_amount": remove_decimals(amount),
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def transfer(self, to: ChecksumAddress, value: Uint256) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.transfer,
            "func_args": {
                "to": to,
                "value": remove_decimals(value),
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def transfer_from(
        self, from_: ChecksumAddress, to: ChecksumAddress, value: Uint256
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.transferFrom,
            "func_args": {
                "from": from_,
                "to": to,
                "value": remove_decimals(value),
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def transfer_with_authorization(
        self,
        from_: ChecksumAddress,
        to: ChecksumAddress,
        value: Uint256,
        valid_after: Uint256,
        valid_before: Uint256,
        nonce: Bytes32,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.transferWithAuthorization,
            "func_args": {
                "from": from_,
                "to": to,
                "value": remove_decimals(value),
                "validAfter": valid_after,
                "validBefore": valid_before,
                "nonce": nonce,
                "v": v,
                "r": r,
                "s": s,
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def receive_with_authorization(
        self,
        from_: ChecksumAddress,
        to: ChecksumAddress,
        value: Uint256,
        valid_after: Uint256,
        valid_before: Uint256,
        nonce: Bytes32,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.receiveWithAuthorization,
            "func_args": {
                "from": from_,
                "to": to,
                "value": remove_decimals(value),
                "validAfter": valid_after,
                "validBefore": valid_before,
                "nonce": nonce,
                "v": v,
                "r": r,
                "s": s,
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def cancel_authorization(
        self,
        authorizer: ChecksumAddress,
        nonce: Bytes32,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.cancelAuthorization,
            "func_args": {
                "authorizer": authorizer,
                "nonce": nonce,
                "v": v,
                "r": r,
                "s": s,
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def approve(self, spender: ChecksumAddress, value: Uint256) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.approve,
            "func_args": {
                "spender": spender,
                "value": remove_decimals(value),
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def increase_allowance(
        self, spender: ChecksumAddress, increment: Uint256
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.increaseAllowance,
            "func_args": {
                "spender": spender,
                "increment": remove_decimals(increment),
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def decrease_allowance(
        self, spender: ChecksumAddress, decrement: Uint256
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.decreaseAllowance,
            "func_args": {
                "spender": spender,
                "decrement": remove_decimals(decrement),
            },
        }

        return self.__transact(**tx_args)

    @validate_call
    def permit(
        self,
        owner: ChecksumAddress,
        spender: ChecksumAddress,
        value: Uint256,
        deadline: Uint256,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        tx_args: TransactionArgs = {
            "contract_func": self.contract.functions.permit,
            "func_args": {
                "owner": owner,
                "spender": spender,
                "value": remove_decimals(value),
                "deadline": deadline,
                "v": v,
                "r": r,
                "s": s,
            },
        }

        return self.__transact(**tx_args)



from .types import ChainName


class JpycSdkError(Exception):
    """A base class for any errors related to JPYC SDK."""

    def __init__(self, code: int, message: str) -> None:
        self.code = code
        """int: Custom error code"""
        self.message = message
        """str: Error message"""
        super().__init__(f"\nError Code: {code}\nMessage: {message}")


#################
# Config Errors #
#################


class NetworkNotSupported(JpycSdkError):
    """Raised when the specified network is not supported by the SDK.

    Attributes:
        chain_name (str): Chain name
        network_name (str): Network name
    """

    code = 100

    def __init__(self, chain_name: ChainName, network_name: str) -> None:
        from .chains import enumerate_supported_networks

        super().__init__(
            code=NetworkNotSupported.code,
            message=f"Network '{chain_name}/{network_name}' is not supported. "
            f"Supported networks are: {enumerate_supported_networks()}",
        )


class AccountNotInitialized(JpycSdkError):
    """Raised when account is not initialized or hoisted to web3 instance."""

    code = 101

    def __init__(self) -> None:
        super().__init__(
            code=AccountNotInitialized.code,
            message="Account is not initialized.",
        )


#####################
# Validation Errors #
#####################
class InvalidChecksumAddress(JpycSdkError, TypeError):
    """Raised when the given address is not a valid checksum address."""

    code = 200

    def __init__(self, message_: str) -> None:
        super().__init__(
            code=InvalidChecksumAddress.code,
            message=f"Invalid checksum address: {message_}. "
            f"Address must be compatible with EIP55.",
        )


class InvalidUint8(JpycSdkError, TypeError):
    """Raised when the given integer is not a valid uint8."""

    code = 201

    def __init__(self, message_: str) -> None:
        super().__init__(
            code=InvalidUint8.code,
            message=f"Invalid uint8: {message_}. Integer must be between 0 ~ 2^8 - 1.",
        )


class InvalidUint256(JpycSdkError, TypeError):
    """Raised when the given integer is not a valid uint256."""

    code = 202

    def __init__(self, message_: str) -> None:
        super().__init__(
            code=InvalidUint256.code,
            message=f"Invalid uint256: {message_}. "
            f"Integer must be between 0 ~ 2^256 - 1.",
        )


class InvalidBytes32(JpycSdkError, TypeError):
    """Raised when the given byte string is not a valid bytes32."""

    code = 203

    def __init__(self, message_: str) -> None:
        super().__init__(
            code=InvalidBytes32.code,
            message=f"Invalid bytes32: {message_}.",
        )


class InvalidRpcEndpoint(JpycSdkError, TypeError):
    """Raised when the given string is not a valid RPC endpoint."""

    code = 204

    def __init__(self, message_: str) -> None:
        super().__init__(
            code=InvalidRpcEndpoint.code,
            message=f"Invalid RPC endpoint: {message_}.",
        )


######################
# Transaction Errors #
######################


class TransactionSimulationFailed(JpycSdkError):
    """Raised when transaction simulation fails.

    Attributes:
        message (str): Error message
    """

    code = 300

    def __init__(self, message_: str) -> None:
        super().__init__(
            code=TransactionSimulationFailed.code,
            message=f"Failed to simulate a transaction locally: {message_}",
        )


class TransactionFailed(JpycSdkError):
    """Raised when transaction fails.

    Attributes:
        message (str): Error message
    """

    code = 301

    def __init__(self, message_: str) -> None:
        super().__init__(
            code=TransactionFailed.code,
            message=f"Transaction failed: {message_}",
        )



from typing import Final

########
# Math #
########

UINT_MIN: Final[int] = 0
"""int: Minimum value of uint."""
UINT8_MAX: Final[int] = 255
"""int: Maximum value of uint8."""
UINT256_MAX: Final[int] = (
    115792089237316195423570985008687907853269984665640564039457584007913129639935
)
"""int: Maximum value of uint256."""

####################
# Middleware Names #
####################

POA_MIDDLEWARE: Final[str] = "poa_middleware"
"""str: Name of POA compatibility middleware."""
SIGN_MIDDLEWARE: Final[str] = "sign_middleware"
"""str: Name of auto-signature middleware."""



from .addresses import get_proxy_address
from .artifacts import (
    get_artifacts,
    resolve_artifacts_file_path,
)
from .chains import (
    SUPPORTED_CHAINS,
    enumerate_supported_networks,
    get_default_rpc_endpoint,
)
from .constants import (
    POA_MIDDLEWARE,
    SIGN_MIDDLEWARE,
    UINT8_MAX,
    UINT256_MAX,
    UINT_MIN,
)
from .currencies import (
    remove_decimals,
    restore_decimals,
)
from .errors import (
    AccountNotInitialized,
    InvalidBytes32,
    InvalidChecksumAddress,
    InvalidUint8,
    InvalidUint256,
    NetworkNotSupported,
    TransactionFailed,
    TransactionSimulationFailed,
)
from .types import (
    ArtifactType,
    ChainMetadata,
    ChainName,
    ContractVersion,
)
from .validators import (
    Bytes32,
    ChecksumAddress,
    RpcEndpoint,
    Uint8,
    Uint256,
)

__all__ = [
    # addresses
    "get_proxy_address",
    # artifacts
    "get_artifacts",
    "resolve_artifacts_file_path",
    # chains
    "enumerate_supported_networks",
    "get_default_rpc_endpoint",
    "SUPPORTED_CHAINS",
    # constants
    "POA_MIDDLEWARE",
    "SIGN_MIDDLEWARE",
    "UINT_MIN",
    "UINT256_MAX",
    "UINT8_MAX",
    # currencies
    "remove_decimals",
    "restore_decimals",
    # errors
    "AccountNotInitialized",
    "InvalidBytes32",
    "InvalidChecksumAddress",
    "InvalidUint8",
    "InvalidUint256",
    "NetworkNotSupported",
    "TransactionFailed",
    "TransactionSimulationFailed",
    # types
    "ArtifactType",
    "ChainMetadata",
    "ChainName",
    "ContractVersion",
    # validators
    "Bytes32",
    "ChecksumAddress",
    "RpcEndpoint",
    "Uint256",
    "Uint8",
]



from typing import Final

from eth_typing import (
    ChecksumAddress,
    HexAddress,
    HexStr,
)
from web3 import Web3
from web3.constants import ADDRESS_ZERO

from .types import ContractVersion

####################################
# Address-related helper functions #
####################################


def calc_checksum_address(address: str) -> ChecksumAddress:
    """Calculates checksum address.

    Args:
        address (str): Address string

    Returns:
        ChecksumAddress: Checksum address
    """
    return ChecksumAddress(HexAddress(HexStr(Web3.to_checksum_address(address))))


def is_valid_address(address: str) -> bool:
    """Checks validity of address.

    Args:
        address (str): Address string

    Returns:
        bool: True if valid, false otherwise
    """
    return Web3.is_checksum_address(address)


def get_proxy_address(contract_version: ContractVersion) -> ChecksumAddress:
    """Get proxy address from the specified version.

    Note:
        Default address should be the address of the latest version \
        (e.g., v2 as of May 2025).

    Args:
        contract_version (ContractVersion): Contract version

    Returns:
        ChecksumAddress: Checksum address of proxy contract
    """
    match contract_version:
        case "2":
            return V2_PROXY_ADDRESS


######################
# Constant addresses #
######################

ZERO_ADDRESS: Final[ChecksumAddress] = calc_checksum_address(str(ADDRESS_ZERO))
"""ChecksumAddress: Zero address."""
V2_PROXY_ADDRESS: Final[ChecksumAddress] = calc_checksum_address(
    "0x431D5dfF03120AFA4bDf332c61A6e1766eF37BDB"
)
"""ChecksumAddress: JPYCv2 address."""



from importlib.metadata import version

from .client import SdkClient
from .jpyc import JPYC

__version__ = version("jpyc-core-sdk")
__all__ = [
    # client
    "SdkClient",
    # jpyc
    "JPYC",
    # utils
    "utils",
]



from decimal import Decimal

from web3 import Web3

from .validators import Uint256


def remove_decimals(value: Uint256 | Decimal) -> Uint256:
    """Remove decimals.

    Args:
        value (Uint256 | Decimal): Value in ether

    Returns:
        Uint256: Value in wei
    """
    return Web3.to_wei(value, "ether")


def restore_decimals(func):  # type: ignore[no-untyped-def]
    """Decorator to restore decimals."""

    def wrapper(*args, **kwargs):  # type: ignore[no-untyped-def]
        result = func(*args, **kwargs)
        return Web3.from_wei(result, "ether")

    return wrapper



from typing import Final

from .errors import NetworkNotSupported
from .types import ChainMetadata, ChainName
from .validators import RpcEndpoint

##################
# Chain Metadata #
##################

SUPPORTED_CHAINS: Final[ChainMetadata] = {
    "ethereum": {
        "mainnet": {
            "id": 1,
            "name": "Ethereum Mainnet",
            "rpc_endpoints": ["https://ethereum-rpc.publicnode.com"],
        },
        "sepolia": {
            "id": 11155111,
            "name": "Ethereum Sepolia Testnet",
            "rpc_endpoints": ["https://ethereum-sepolia-rpc.publicnode.com"],
        },
    },
    "polygon": {
        "mainnet": {
            "id": 137,
            "name": "Polygon Mainnet",
            "rpc_endpoints": ["https://polygon-rpc.com"],
        },
        "amoy": {
            "id": 80002,
            "name": "Polygon Amoy Testnet",
            "rpc_endpoints": ["https://rpc-amoy.polygon.technology"],
        },
    },
    "gnosis": {
        "mainnet": {
            "id": 100,
            "name": "Gnosis Chain",
            "rpc_endpoints": ["https://rpc.gnosischain.com"],
        },
        "chiado": {
            "id": 10200,
            "name": "Gnosis Chiado Testnet",
            "rpc_endpoints": ["https://rpc.chiadochain.net"],
        },
    },
    "avalanche": {
        "mainnet": {
            "id": 43114,
            "name": "Avalanche C-Chain",
            "rpc_endpoints": ["https://api.avax.network/ext/bc/C/rpc"],
        },
        "fuji": {
            "id": 43113,
            "name": "Avalanche Fuji Testnet",
            "rpc_endpoints": ["https://api.avax-test.network/ext/bc/C/rpc"],
        },
    },
    "astar": {
        "mainnet": {
            "id": 592,
            "name": "Astar Network",
            "rpc_endpoints": ["https://astar.public.blastapi.io"],
        },
    },
    "shiden": {
        "mainnet": {
            "id": 336,
            "name": "Shiden Network",
            "rpc_endpoints": ["https://shiden.public.blastapi.io"],
        },
    },
    "localhost": {
        "devnet": {
            "id": 31337,
            "name": "Localhost Network",
            "rpc_endpoints": ["http://127.0.0.1:8545/"],
        },
    },
}
"""ChainMetadata: Supported chains & networks."""

##################################
# Chain-related helper functions #
##################################


def enumerate_supported_networks() -> str:
    """Enumerate all the supported networks.

    Returns:
        str: supported networks
    """
    return ", ".join(
        f"'{chain}' => {list(networks.keys())}"
        for chain, networks in SUPPORTED_CHAINS.items()
    )


def is_supported_network(
    chain_name: ChainName | None, network_name: str | None
) -> bool:
    """Check if the specified network is supported by the SDK.

    Args:
        chain_name (ChainName, optional): Chain name
        network_name (str, optional): Network name

    Returns:
        bool: True if supported, false otherwise
    """
    return (
        chain_name in SUPPORTED_CHAINS and network_name in SUPPORTED_CHAINS[chain_name]
    )


def get_default_rpc_endpoint(
    chain_name: ChainName | None, network_name: str | None
) -> RpcEndpoint:
    """Get the default RPC endpoint for the specified network.

    Args:
        chain_name (ChainName, optional): Chain name
        network_name (str, optional): Network name

    Returns:
        RpcEndpoint: RPC endpoint

    Raises:
        NetworkNotSupported: If the specified network is not supported by the SDK
    """
    if not is_supported_network(chain_name, network_name):
        raise NetworkNotSupported(chain_name, network_name)  # type: ignore[arg-type]

    return SUPPORTED_CHAINS[chain_name][network_name]["rpc_endpoints"][0]  # type: ignore[index]



from eth_account.signers.local import LocalAccount
from pydantic import validate_call
from web3 import Account, HTTPProvider, Web3
from web3._utils.empty import empty
from web3.middleware import (
    ExtraDataToPOAMiddleware,
    SignAndSendRawMiddlewareBuilder,
)

from .interfaces import ISdkClient
from .utils.chains import get_default_rpc_endpoint
from .utils.constants import (
    POA_MIDDLEWARE,
    SIGN_MIDDLEWARE,
)
from .utils.errors import AccountNotInitialized
from .utils.types import ChainName
from .utils.validators import (
    Bytes32,
    ChecksumAddress,
    RpcEndpoint,
)


class SdkClient(ISdkClient):
    """SDK client."""

    @validate_call
    def __init__(
        self,
        chain_name: ChainName | None = None,
        network_name: str | None = None,
        rpc_endpoint: RpcEndpoint | None = None,
        private_key: Bytes32 | None = None,
    ) -> None:
        """Constructor that initializes SDK client.

        Notes:
            - Either `chain_name` & `network_name` parameters\
              or `rpc_endpoint` parameter are required.
            - This constructor prioritizes `rpc_endpoint` parameter over\
            `chain_name` & `network_name` parameters when configuring `rpc_endpoint`.

        Args:
            chain_name (str, optional): Chain name
            network_name (str, optional): Network name
            rpc_endpoint (RpcEndpoint, optional): RPC endpoint
            private_key (Bytes32, optional): private key of EOA

        Raises:
            InvalidBytes32: If the supplied `private_key` is not in a valid form
            InvalidRpcEndpoint: If the supplied `rpc_endpoint` is not in a valid form
            NetworkNotSupported: If the specified network is not supported by the SDK
            ValidationError: If pydantic validation fails
        """
        rpc_endpoint = (
            rpc_endpoint
            if rpc_endpoint is not None
            else get_default_rpc_endpoint(chain_name, network_name)
        )
        account = Account.from_key(private_key) if private_key is not None else None
        w3 = self.__configure_w3(
            rpc_endpoint=rpc_endpoint,
            account=account,
        )

        self.w3 = w3
        """Web3: Configured web3 instance"""
        self.rpc_endpoint = rpc_endpoint
        """str: RPC endpoint"""
        self.account = account
        """LocalAccount | None: Account instance"""

    @staticmethod
    def __configure_w3(
        rpc_endpoint: RpcEndpoint,
        account: LocalAccount | None = None,
    ) -> Web3:
        """Configure a web3 instance.

        Args:
            rpc_endpoint (RpcEndpoint): RPC endpoint
            account (LocalAccount, optional): Account instance

        Returns:
            Web3: Configured web3 instance
        """
        w3 = Web3(HTTPProvider(rpc_endpoint))
        w3.middleware_onion.inject(
            ExtraDataToPOAMiddleware,
            name=POA_MIDDLEWARE,
            layer=0,
        )
        if account is not None:
            w3.eth.default_account = account.address
            w3.middleware_onion.inject(
                SignAndSendRawMiddlewareBuilder.build(account),
                name=SIGN_MIDDLEWARE,
                layer=0,
            )
        else:
            w3.eth.default_account = empty

        return w3

    @validate_call
    def set_default_provider(self, chain_name: ChainName, network_name: str) -> Web3:
        self.w3 = self.__configure_w3(
            rpc_endpoint=get_default_rpc_endpoint(chain_name, network_name),
            account=self.account,
        )

        return self.w3

    @validate_call
    def set_custom_provider(self, rpc_endpoint: RpcEndpoint) -> Web3:
        self.w3 = self.__configure_w3(rpc_endpoint=rpc_endpoint, account=self.account)

        return self.w3

    @validate_call
    def set_account(self, private_key: Bytes32 | None) -> LocalAccount | None:
        if private_key is None:
            self.account = None
            self.w3 = self.__configure_w3(
                rpc_endpoint=self.rpc_endpoint,
            )
        else:
            self.account = Account.from_key(private_key)
            self.w3 = self.__configure_w3(
                rpc_endpoint=self.rpc_endpoint, account=self.account
            )

        return self.account

    def get_account_address(self) -> ChecksumAddress:
        if self.account is None:
            raise AccountNotInitialized()

        return self.account.address



import json
from pathlib import Path
from typing import Any

from .types import ArtifactType, ContractVersion


def resolve_artifacts_file_path(contract_version: ContractVersion) -> Path:
    """Resolve the path of artifacts file from the specified contract version.

    Args:
        contract_version (ContractVersion): Contract version

    Returns:
        Path: Absolute path of artifacts file
    """
    path = Path(__file__).parent.parent.joinpath(
        "artifacts", f"v{contract_version}.json"
    )

    return path.absolute()


def get_artifacts(file_path: Path, artifact_type: ArtifactType) -> Any:
    """Get contract artifacts from the specified file path.

    Args:
        file_path (Path): absolute path of artifacts file
        artifact_type (ArtifactType): type of artifacts

    Returns:
        Any: Artifacts of contracts
    """
    with open(file_path) as f:
        return json.load(f)[artifact_type]



from typing import Annotated

from pydantic.functional_validators import AfterValidator
from web3 import Web3

from .constants import UINT8_MAX, UINT256_MAX, UINT_MIN
from .errors import (
    InvalidBytes32,
    InvalidChecksumAddress,
    InvalidRpcEndpoint,
    InvalidUint8,
    InvalidUint256,
)


def validate_checksum_address(address: str) -> str:
    """Checks if the given address is a checksum address.

    Args:
        address (str): Address string

    Returns:
        str: Address string

    Raises:
        InvalidChecksumAddress: If the address is not a valid checksum address
    """
    if not Web3.is_checksum_address(address):
        raise InvalidChecksumAddress(address)

    return address


ChecksumAddress = Annotated[str, AfterValidator(validate_checksum_address)]
"""A type that contains a checksum address. \
See `EIP55 <https://eips.ethereum.org/EIPS/eip-55>`_ for more details.
"""


def validate_uint8(integer: int) -> int:
    """Checks if the given integer is a uint8.

    Args:
        integer (int): Integer

    Returns:
        int: Integer

    Raises:
        InvalidUint8: If the integer is not a valid uint8
    """
    if integer < UINT_MIN or UINT8_MAX < integer:
        raise InvalidUint8(str(integer))

    return integer


Uint8 = Annotated[int, AfterValidator(validate_uint8)]
"""A type that contains a uint8."""


def validate_uint256(integer: int) -> int:
    """Checks if the given integer is a uint256.

    Args:
        integer (int): Integer

    Returns:
        int: Integer

    Raises:
        InvalidUint256: If the integer is not a valid uint256
    """
    if integer < UINT_MIN or UINT256_MAX < integer:
        raise InvalidUint256(str(integer))

    return integer


Uint256 = Annotated[int, AfterValidator(validate_uint256)]
"""A type that contains a uint256."""


def validate_bytes32(string: str) -> str:
    """Checks if the given string is bytes32.

    Args:
        string (str): String

    Returns:
        str: String

    Raises:
        InvalidBytes32: If the string is not a valid bytes32
    """
    try:
        assert string[:2] == "0x"
        assert len(bytes.fromhex(string[2:])) == 32
    except Exception:
        raise InvalidBytes32(string)

    return string


Bytes32 = Annotated[str, AfterValidator(validate_bytes32)]
"""A type that contains a bytes32."""


def validate_rpc_endpoint(string: str) -> str:
    """Checks if the given string is a valid RPC endpoint.

    Args:
        string (str): String

    Returns:
        str: String

    Raises:
        InvalidRpcEndpoint: If the string is not a valid RPC endpoint
    """
    try:
        assert string.startswith("http") is True
    except Exception:
        raise InvalidRpcEndpoint(string)

    return string


RpcEndpoint = Annotated[str, AfterValidator(validate_rpc_endpoint)]
"""A type that contains a rpc endpoint."""



from abc import ABC, abstractmethod

from eth_account.signers.local import LocalAccount
from web3 import Web3

from ..utils.types import ChainName
from ..utils.validators import (
    Bytes32,
    ChecksumAddress,
    RpcEndpoint,
)


class ISdkClient(ABC):
    """Interface of SDK client."""

    @abstractmethod
    def set_default_provider(self, chain_name: ChainName, network_name: str) -> Web3:
        """Set provider using one of the default RPC endpoints.

        Args:
            chain_name (str): Chain name
            network_name (str): Network name

        Returns:
            Web3: Configured web3 instance

        Raises:
            NetworkNotSupported: If the specified network is not supported by the SDK
            ValidationError: If pydantic validation fails
        """
        pass

    @abstractmethod
    def set_custom_provider(self, rpc_endpoint: RpcEndpoint) -> Web3:
        """Set provider using a custom RPC endpoint.

        Args:
            rpc_endpoint (RpcEndpoint): Custom RPC endpoint

        Returns:
            Web3: Configured web3 instance

        Raises:
            InvalidRpcEndpoint: If the supplied `rpc_endpoint` is not in a valid form
        """
        pass

    @abstractmethod
    def set_account(self, private_key: Bytes32 | None) -> LocalAccount | None:
        """Set account with a private key.

        Notes:
            If `private_key` parameter is set to `None`, \
            this method removes `account` from the configured web3 instance.

        Args:
            private_key (Bytes32, optional): Private key of account

        Returns:
            LocalAccount | None: Configured account instance

        Raises:
            InvalidBytes32: If the supplied `private_key` is not in a valid form
        """
        pass

    @abstractmethod
    def get_account_address(self) -> ChecksumAddress:
        """Get address of the configured account.

        Returns:
            ChecksumAddress: Public address of account

        Raises:
            AccountNotInitialized: If account is not initialized
        """
        pass



from abc import ABC, abstractmethod

from ..utils.validators import (
    Bytes32,
    ChecksumAddress,
    Uint8,
    Uint256,
)


class IJPYC(ABC):
    """Interface of JPYC contracts."""

    ##################
    # View functions #
    ##################

    @abstractmethod
    def is_minter(self, account: ChecksumAddress) -> bool:
        """Call `isMinter` function.

        Args:
            account (ChecksumAddress): Account address

        Returns:
            bool: True if `account` is a minter, false otherwise

        Raises:
            InvalidChecksumAddress: If supplied `account` is not in a valid form
        """
        pass

    @abstractmethod
    def minter_allowance(self, minter: ChecksumAddress) -> Uint256:
        """Call `minterAllowance` function.

        Args:
            minter (ChecksumAddress): Minter address

        Returns:
            Uint256: Minter allowance

        Raises:
            InvalidChecksumAddress: If supplied `minter` is not in a valid form
        """
        pass

    @abstractmethod
    def total_supply(self) -> Uint256:
        """Call `totalSupply` function.

        Returns:
            Uint256: Total supply of tokens
        """
        pass

    @abstractmethod
    def balance_of(self, account: ChecksumAddress) -> Uint256:
        """Call `balanceOf` function.

        Args:
            account (ChecksumAddress): Account address

        Returns:
            Uint256: Account balance

        Raises:
            InvalidChecksumAddress: If supplied `account` is not in a valid form
        """
        pass

    @abstractmethod
    def allowance(self, owner: ChecksumAddress, spender: ChecksumAddress) -> Uint256:
        """Call `allowance` function.

        Args:
            owner (ChecksumAddress): Owner address
            spender (ChecksumAddress): Spender address

        Returns:
            Uint256: Allowance of spender over owner's tokens

        Raises:
            InvalidChecksumAddress: If supplied `owner` or `spender`\
                                    is not in a valid form
        """
        pass

    @abstractmethod
    def nonces(self, owner: ChecksumAddress) -> Uint256:
        """Call `nonces` function.

        Args:
            owner (ChecksumAddress): Owner address

        Returns:
            Uint256: Nonce for EIP2612's `permit`.

        Raises:
            InvalidChecksumAddress: If supplied `owner` is not in a valid form
        """
        pass

    ######################
    # Mutation functions #
    ######################

    @abstractmethod
    def configure_minter(
        self, minter: ChecksumAddress, minter_allowed_amount: Uint256
    ) -> Bytes32:
        """Call `configureMinter` function.

        Args:
            minter (ChecksumAddress): Minter address
            minter_allowed_amount (Uint256): Minter allowance

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidChecksumAddress: If supplied `minter` is not in a valid form
            InvalidUint256: If supplied `minter_allowed_amount` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def mint(self, to: ChecksumAddress, amount: Uint256) -> Bytes32:
        """Call `mint` function.

        Args:
            to (ChecksumAddress): Receiver address
            amount (Uint256): Amount of tokens to mint

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidChecksumAddress: If supplied `to` is not in a valid form
            InvalidUint256: If supplied `amount` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def transfer(self, to: ChecksumAddress, value: Uint256) -> Bytes32:
        """Call `transfer` function.

        Args:
            to (ChecksumAddress): Receiver address
            value (Uint256): Amount of tokens to transfer

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidChecksumAddress: If supplied `to` is not in a valid form
            InvalidUint256: If supplied `value` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def transfer_from(
        self, from_: ChecksumAddress, to: ChecksumAddress, value: Uint256
    ) -> Bytes32:
        """Call `transferFrom` function.

        Args:
            from_ (ChecksumAddress): Owner address
            to (ChecksumAddress): Receiver address
            value (Uint256): Amount of tokens to transfer

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidChecksumAddress: If supplied `from_` or `to` is not in a valid form
            InvalidUint256: If supplied `value` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def transfer_with_authorization(
        self,
        from_: ChecksumAddress,
        to: ChecksumAddress,
        value: Uint256,
        valid_after: Uint256,
        valid_before: Uint256,
        nonce: Bytes32,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        """Call `transferWithAuthorization` function.

        Args:
            from_ (ChecksumAddress): Owner address
            to (ChecksumAddress): Receiver allowance
            value (Uint256): Amount of tokens to transfer
            valid_after (Uint256): Unix time when transaction becomes valid
            valid_before (Uint256): Unix time when transaction becomes invalid
            nonce (Bytes32): Unique nonce
            v (Uint8): v of ECDSA
            r (Bytes32): r of ECDSA
            s (Bytes32): s of ECDSA

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidBytes32: If supplied `nonce` or `r` or `s` is not in a valid form
            InvalidChecksumAddress: If supplied `from_` or `to` is not in a valid form
            InvalidUint256: If supplied `value` or `valid_after` or `valid_before`\
                            is not in a valid form
            InvalidUint8: If supplied `v` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def receive_with_authorization(
        self,
        from_: ChecksumAddress,
        to: ChecksumAddress,
        value: Uint256,
        valid_after: Uint256,
        valid_before: Uint256,
        nonce: Bytes32,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        """Call `receiveWithAuthorization` function.

        Args:
            from_ (ChecksumAddress): Owner address
            to (ChecksumAddress): Receiver allowance
            value (Uint256): Amount of tokens to transfer
            valid_after (Uint256): Unix time when transaction becomes valid
            valid_before (Uint256): Unix time when transaction becomes invalid
            nonce (Bytes32): Unique nonce
            v (Uint8): v of ECDSA
            r (Bytes32): r of ECDSA
            s (Bytes32): s of ECDSA

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidBytes32: If supplied `nonce` or `r` or `s` is not in a valid form
            InvalidChecksumAddress: If supplied `from_` or `to` is not in a valid form
            InvalidUint256: If supplied `value` or `valid_after` or `valid_before`\
                            is not in a valid form
            InvalidUint8: If supplied `v` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def cancel_authorization(
        self,
        authorizer: ChecksumAddress,
        nonce: Bytes32,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        """Call `cancelAuthorization` function.

        Args:
            authorizer (ChecksumAddress): Owner address
            nonce (Bytes32): Unique nonce
            v (Uint8): v of ECDSA
            r (Bytes32): r of ECDSA
            s (Bytes32): s of ECDSA

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidBytes32: If supplied `nonce` or `r` or `s` is not in a valid form
            InvalidChecksumAddress: If supplied `authorizer` is not in a valid form
            InvalidUint8: If supplied `v` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def approve(self, spender: ChecksumAddress, value: Uint256) -> Bytes32:
        """Call `approve` function.

        Args:
            spender (ChecksumAddress): Spender address
            value (Uint256): Amount of allowance

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidChecksumAddress: If supplied `spender` is not in a valid form
            InvalidUint256: If supplied `value` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def increase_allowance(
        self, spender: ChecksumAddress, increment: Uint256
    ) -> Bytes32:
        """Call `increaseAllowance` function.

        Args:
            spender (ChecksumAddress): Spender address
            increment (Uint256): Amount of allowance to increase

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidChecksumAddress: If supplied `spender` is not in a valid form
            InvalidUint256: If supplied `increment` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def decrease_allowance(
        self, spender: ChecksumAddress, decrement: Uint256
    ) -> Bytes32:
        """Call `decreaseAllowance` function.

        Args:
            spender (ChecksumAddress): Spender address
            decrement (Uint256): Amount of allowance to decrease

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidChecksumAddress: If supplied `spender` is not in a valid form
            InvalidUint256: If supplied `decrement` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass

    @abstractmethod
    def permit(
        self,
        owner: ChecksumAddress,
        spender: ChecksumAddress,
        value: Uint256,
        deadline: Uint256,
        v: Uint8,
        r: Bytes32,
        s: Bytes32,
    ) -> Bytes32:
        """Call `permit` function.

        Args:
            owner (ChecksumAddress): Owner address
            spender (ChecksumAddress): Spender address
            value (Uint256): Amount of allowance
            deadline (Uint256): Unix time when transaction becomes invalid
            v (Uint8): v of ECDSA
            r (Bytes32): r of ECDSA
            s (Bytes32): s of ECDSA

        Returns:
            Bytes32: Transaction hash

        Raises:
            AccountNotInitialized: If account is not initialized
            InvalidBytes32: If supplied `r` or `s` is not in a valid form
            InvalidChecksumAddress: If supplied `owner` or `spender`\
                                    is not in a valid form
            InvalidUint256: If supplied `value` or `deadline` is not in a valid form
            InvalidUint8: If supplied `v` is not in a valid form
            TransactionFailed: If transaction fails
            TransactionSimulationFailed: If transaction simulation fails
        """
        pass



from typing import Literal, TypedDict

from web3.contract.contract import ContractFunction

##########
# Chains #
##########

type ChainName = Literal[
    "ethereum",
    "polygon",
    "gnosis",
    "avalanche",
    "astar",
    "shiden",
    "localhost",
]


class NetworkMetadata(TypedDict):
    id: int
    name: str
    rpc_endpoints: list[str]


type ChainMetadata = dict[ChainName, dict[str, NetworkMetadata]]
"""A type that contains metadata of chains."""

#############
# Contracts #
#############

type ContractVersion = Literal["2"]
"""A type that contains available contract versions."""
type ArtifactType = Literal["abi", "bytecode"]
"""A type that contains types of contract artifacts."""

################
# Transactions #
################


class TransactionArgs(TypedDict):
    contract_func: ContractFunction
    func_args: dict[str, object]
